@@ -1,32 +1,28 @@
-# -*- coding: utf-8 -*-
-"""
+"""Treebeard exceptions"""
 
-    treebeard.exceptions
-    --------------------
-
-    Exceptions
-
-    :copyright: 2008 by Gustavo Picon
-    :license: Apache License 2.0
-
-"""
 
 class InvalidPosition(Exception):
-    """
-    Raised when passing an invalid pos value
-    """
+    """Raised when passing an invalid pos value"""
+
 
 class InvalidMoveToDescendant(Exception):
+    """Raised when attemping to move a node to one of it's descendants."""
+
+
+class NodeAlreadySaved(Exception):
     """
-    Raised when attemping to move a node to one of it's descendants.
+    Raised when attempting to add a node which is already saved to the
+    database.
     """
 
+
 class MissingNodeOrderBy(Exception):
     """
     Raised when an operation needs a missing
     :attr:`~treebeard.MP_Node.node_order_by` attribute
     """
 
+
 class PathOverflow(Exception):
     """
     Raised when trying to add or move a node to a position where no more nodes
