@@ -1,17 +1,21 @@
-"Models and base API"
+"""Models and base API"""
 
+import sys
 import operator
 
-from django import VERSION as DJANGO_VERSION
+if sys.version_info >= (3, 0):
+    from functools import reduce
+
 from django.db.models import Q
-from django.db import models, transaction
-from django.conf import settings
+from django.db import models, transaction, router, connections
 
 from treebeard.exceptions import InvalidPosition, MissingNodeOrderBy
 
 
 class Node(models.Model):
-    "Node class"
+    """Node class"""
+
+    _db_connection = None
 
     @classmethod
     def add_root(cls, **kwargs):  # pragma: no cover
@@ -23,11 +27,42 @@ class Node(models.Model):
 
         :param \*\*kwargs: object creation data that will be passed to the
             inherited Node model
+        :param instance: Instead of passing object creation data, you can
+            pass an already-constructed (but not yet saved) model instance to
+            be inserted into the tree.
 
         :returns: the created node object. It will be save()d by this method.
+
+        :raise NodeAlreadySaved: when the passed ``instance`` already exists
+            in the database
         """
         raise NotImplementedError
 
+    @classmethod
+    def get_foreign_keys(cls):
+        """Get foreign keys and models they refer to, so we can pre-process
+        the data for load_bulk
+        """
+        foreign_keys = {}
+        for field in cls._meta.fields:
+            if (
+                field.get_internal_type() == 'ForeignKey' and
+                field.name != 'parent'
+            ):
+                foreign_keys[field.name] = field.rel.to
+        return foreign_keys
+
+    @classmethod
+    def _process_foreign_keys(cls, foreign_keys, node_data):
+        """For each foreign key try to load the actual object so load_bulk
+        doesn't fail trying to load an int where django expects a
+        model instance
+        """
+        for key in foreign_keys.keys():
+            if key in node_data:
+                node_data[key] = foreign_keys[key].objects.get(
+                    pk=node_data[key])
+
     @classmethod
     def load_bulk(cls, bulk_data, parent=None, keep_ids=False):
         """
@@ -55,7 +90,7 @@ class Node(models.Model):
 
         :param keep_ids:
 
-            If enabled, lads the nodes with the same id that are given in the
+            If enabled, loads the nodes with the same id that are given in the
             structure. Will error if there are nodes without id info or if the
             ids are already used.
 
@@ -67,23 +102,27 @@ class Node(models.Model):
         added = []
         # stack of nodes to analize
         stack = [(parent, node) for node in bulk_data[::-1]]
+        foreign_keys = cls.get_foreign_keys()
+
         while stack:
             parent, node_struct = stack.pop()
             # shallow copy of the data strucure so it doesn't persist...
             node_data = node_struct['data'].copy()
+            cls._process_foreign_keys(foreign_keys, node_data)
             if keep_ids:
                 node_data['id'] = node_struct['id']
             if parent:
                 node_obj = parent.add_child(**node_data)
             else:
                 node_obj = cls.add_root(**node_data)
-            added.append(node_obj.id)
+            added.append(node_obj.pk)
             if 'children' in node_struct:
                 # extending the stack with the current node as the parent of
                 # the new nodes
-                stack.extend([(node_obj, node) \
-                    for node in node_struct['children'][::-1]])
-        transaction.commit_unless_managed()
+                stack.extend([
+                    (node_obj, node)
+                    for node in node_struct['children'][::-1]
+                ])
         return added
 
     @classmethod
@@ -101,19 +140,23 @@ class Node(models.Model):
             Stores the id value (primary key) of every node. Enabled by
             default.
 
-        :returns: A python data structure, describen with detail in
+        :returns: A python data structure, described with detail in
                   :meth:`load_bulk`
         """
         raise NotImplementedError
 
     @classmethod
     def get_root_nodes(cls):  # pragma: no cover
-        ":returns: A queryset containing the root nodes in the tree."
+        """:returns: A queryset containing the root nodes in the tree."""
         raise NotImplementedError
 
     @classmethod
     def get_first_root_node(cls):
-        ":returns: The first root node in the tree or ``None`` if it is empty"
+        """
+        :returns:
+
+            The first root node in the tree or ``None`` if it is empty.
+        """
         try:
             return cls.get_root_nodes()[0]
         except IndexError:
@@ -121,7 +164,11 @@ class Node(models.Model):
 
     @classmethod
     def get_last_root_node(cls):
-        ":returns: The last root node in the tree or ``None`` if it is empty"
+        """
+        :returns:
+
+            The last root node in the tree or ``None`` if it is empty.
+        """
         try:
             return cls.get_root_nodes().reverse()[0]
         except IndexError:
@@ -129,7 +176,7 @@ class Node(models.Model):
 
     @classmethod
     def find_problems(cls):  # pragma: no cover
-        "Checks for problems in the tree structure."
+        """Checks for problems in the tree structure."""
         raise NotImplementedError
 
     @classmethod
@@ -143,8 +190,10 @@ class Node(models.Model):
     @classmethod
     def get_tree(cls, parent=None):
         """
-        :returns: A list of nodes ordered as DFS, including the parent. If
-                  no parent is given, the entire tree is returned.
+        :returns:
+
+            A list of nodes ordered as DFS, including the parent. If
+            no parent is given, the entire tree is returned.
         """
         raise NotImplementedError
 
@@ -164,9 +213,6 @@ class Node(models.Model):
             A `list` (**NOT** a Queryset) of node objects with an extra
             attribute: `descendants_count`.
         """
-
-        # this is the slowest possible implementation, subclasses should do
-        # better
         if parent is None:
             qset = cls.get_root_nodes()
         else:
@@ -177,47 +223,56 @@ class Node(models.Model):
         return nodes
 
     def get_depth(self):  # pragma: no cover
-        ":returns: the depth (level) of the node"
+        """:returns: the depth (level) of the node"""
         raise NotImplementedError
 
     def get_siblings(self):  # pragma: no cover
         """
-        :returns: A queryset of all the node's siblings, including the node
+        :returns:
+
+            A queryset of all the node's siblings, including the node
             itself.
         """
         raise NotImplementedError
 
     def get_children(self):  # pragma: no cover
-        ":returns: A queryset of all the node's children"
+        """:returns: A queryset of all the node's children"""
         raise NotImplementedError
 
     def get_children_count(self):
-        ":returns: The number of the node's children"
-
-        # this is the last resort, subclasses of Node should implement this in
-        # a efficient way.
+        """:returns: The number of the node's children"""
         return self.get_children().count()
 
     def get_descendants(self):
         """
-        :returns: A queryset of all the node's descendants, doesn't
+        :returns:
+
+            A queryset of all the node's descendants, doesn't
             include the node itself (some subclasses may return a list).
         """
         raise NotImplementedError
 
     def get_descendant_count(self):
-        ":returns: the number of descendants of a node."
+        """:returns: the number of descendants of a node."""
         return self.get_descendants().count()
 
     def get_first_child(self):
-        ":returns: The leftmost node's child, or None if it has no children."
+        """
+        :returns:
+
+            The leftmost node's child, or None if it has no children.
+        """
         try:
             return self.get_children()[0]
         except IndexError:
             return None
 
     def get_last_child(self):
-        ":returns: The rightmost node's child, or None if it has no children."
+        """
+        :returns:
+
+            The rightmost node's child, or None if it has no children.
+        """
         try:
             return self.get_children().reverse()[0]
         except IndexError:
@@ -225,24 +280,29 @@ class Node(models.Model):
 
     def get_first_sibling(self):
         """
-        :returns: The leftmost node's sibling, can return the node itself if
+        :returns:
+
+            The leftmost node's sibling, can return the node itself if
             it was the leftmost sibling.
         """
         return self.get_siblings()[0]
 
     def get_last_sibling(self):
         """
-        :returns: The rightmost node's sibling, can return the node itself if
+        :returns:
+
+            The rightmost node's sibling, can return the node itself if
             it was the rightmost sibling.
         """
         return self.get_siblings().reverse()[0]
 
     def get_prev_sibling(self):
         """
-        :returns: The previous node's sibling, or None if it was the leftmost
+        :returns:
+
+            The previous node's sibling, or None if it was the leftmost
             sibling.
         """
-
         siblings = self.get_siblings()
         ids = [obj.pk for obj in siblings]
         if self.pk in ids:
@@ -252,7 +312,9 @@ class Node(models.Model):
 
     def get_next_sibling(self):
         """
-        :returns: The next node's sibling, or None if it was the rightmost
+        :returns:
+
+            The next node's sibling, or None if it was the rightmost
             sibling.
         """
         siblings = self.get_siblings()
@@ -264,14 +326,14 @@ class Node(models.Model):
 
     def is_sibling_of(self, node):
         """
-        :returns: ``True`` if the node if a sibling of another node given as an
+        :returns: ``True`` if the node is a sibling of another node given as an
             argument, else, returns ``False``
 
         :param node:
 
             The node that will be checked as a sibling
         """
-        return len(self.get_siblings().filter(pk__in=[node.pk])) > 0
+        return self.get_siblings().filter(pk=node.pk).exists()
 
     def is_child_of(self, node):
         """
@@ -282,11 +344,11 @@ class Node(models.Model):
 
             The node that will be checked as a parent
         """
-        return len(node.get_children().filter(pk__in=[self.pk])) > 0
+        return node.get_children().filter(pk=self.pk).exists()
 
     def is_descendant_of(self, node):  # pragma: no cover
         """
-        :returns: ``True`` if the node if a descendant of another node given
+        :returns: ``True`` if the node is a descendant of another node given
             as an argument, else, returns ``False``
 
         :param node:
@@ -306,8 +368,14 @@ class Node(models.Model):
 
             Object creation data that will be passed to the inherited Node
             model
+        :param instance: Instead of passing object creation data, you can
+            pass an already-constructed (but not yet saved) model instance to
+            be inserted into the tree.
 
         :returns: The created node object. It will be save()d by this method.
+
+        :raise NodeAlreadySaved: when the passed ``instance`` already exists
+            in the database
         """
         raise NotImplementedError
 
@@ -332,6 +400,9 @@ class Node(models.Model):
 
             Object creation data that will be passed to the inherited
             Node model
+        :param instance: Instead of passing object creation data, you can
+            pass an already-constructed (but not yet saved) model instance to
+            be inserted into the tree.
 
         :returns:
 
@@ -342,24 +413,28 @@ class Node(models.Model):
            ``pos`` parm wasn't ``sorted-sibling``
         :raise MissingNodeOrderBy: when passing ``sorted-sibling`` as ``pos``
            and the :attr:`node_order_by` attribute is missing
+        :raise NodeAlreadySaved: when the passed ``instance`` already exists
+            in the database
         """
         raise NotImplementedError
 
     def get_root(self):  # pragma: no cover
-        ":returns: the root node for the current node object."
+        """:returns: the root node for the current node object."""
         raise NotImplementedError
 
     def is_root(self):
-        ":returns: True if the node is a root node (else, returns False)"
-        return self.get_root() == self
+        """:returns: True if the node is a root node (else, returns False)"""
+        return self.get_root().pk == self.pk
 
     def is_leaf(self):
-        ":returns: True if the node is a leaf node (else, returns False)"
-        return self.get_children_count() == 0
+        """:returns: True if the node is a leaf node (else, returns False)"""
+        return not self.get_children().exists()
 
     def get_ancestors(self):  # pragma: no cover
         """
-        :returns: A queryset containing the current node object's ancestors,
+        :returns:
+
+            A queryset containing the current node object's ancestors,
             starting by the root node and descending to the parent.
             (some subclasses may return a list)
         """
@@ -406,9 +481,10 @@ class Node(models.Model):
               the ``target`` node according to the value of
               :attr:`node_order_by`
 
-            .. note:: If no ``pos`` is given the library will use
-                     ``last-sibling``, or ``sorted-sibling`` if
-                     :attr:`node_order_by` is enabled.
+            .. note::
+
+               If no ``pos`` is given the library will use ``last-sibling``,
+               or ``sorted-sibling`` if :attr:`node_order_by` is enabled.
 
         :returns: None
 
@@ -426,69 +502,93 @@ class Node(models.Model):
         raise NotImplementedError
 
     def delete(self):
-        "Removes a node and all it's descendants."
-        self.__class__.objects.filter(id=self.id).delete()
+        """Removes a node and all it's descendants."""
+        self.__class__.objects.filter(pk=self.pk).delete()
 
-    def _fix_add_sibling_opts(self, pos):
-        "prepare the pos variable for the add_sibling method"
+    def _prepare_pos_var(self, pos, method_name, valid_pos, valid_sorted_pos):
         if pos is None:
             if self.node_order_by:
                 pos = 'sorted-sibling'
             else:
                 pos = 'last-sibling'
-        if pos not in ('first-sibling', 'left', 'right', 'last-sibling',
-                       'sorted-sibling'):
+        if pos not in valid_pos:
             raise InvalidPosition('Invalid relative position: %s' % (pos, ))
-        if self.node_order_by and pos != 'sorted-sibling':
-            raise InvalidPosition('Must use %s in add_sibling when'
-                                  ' node_order_by is enabled' % (
-                                  'sorted-sibling', ))
-        if pos == 'sorted-sibling' and not self.node_order_by:
+        if self.node_order_by and pos not in valid_sorted_pos:
+            raise InvalidPosition(
+                'Must use %s in %s when node_order_by is enabled' % (
+                    ' or '.join(valid_sorted_pos), method_name))
+        if pos in valid_sorted_pos and not self.node_order_by:
             raise MissingNodeOrderBy('Missing node_order_by attribute.')
         return pos
 
-    def _fix_move_opts(self, pos):
-        "prepare the pos var for the move method"
-        if pos is None:
-            if self.node_order_by:
-                pos = 'sorted-sibling'
-            else:
-                pos = 'last-sibling'
-        if pos not in ('first-sibling', 'left', 'right', 'last-sibling',
-                       'sorted-sibling', 'first-child', 'last-child',
-                       'sorted-child'):
-            raise InvalidPosition('Invalid relative position: %s' % (pos, ))
-        if self.node_order_by and pos not in ('sorted-child',
-                                              'sorted-sibling'):
-            raise InvalidPosition('Must use %s or %s in add_sibling when'
-                                  ' node_order_by is enabled' % (
-                                  'sorted-sibling', 'sorted-child'))
-        if pos in ('sorted-child', 'sorted-sibling') and \
-                not self.node_order_by:
-            raise MissingNodeOrderBy('Missing node_order_by attribute.')
-        return pos
+    _valid_pos_for_add_sibling = ('first-sibling', 'left', 'right',
+                                  'last-sibling', 'sorted-sibling')
+    _valid_pos_for_sorted_add_sibling = ('sorted-sibling',)
+
+    def _prepare_pos_var_for_add_sibling(self, pos):
+        return self._prepare_pos_var(
+            pos,
+            'add_sibling',
+            self._valid_pos_for_add_sibling,
+            self._valid_pos_for_sorted_add_sibling)
+
+    _valid_pos_for_move = _valid_pos_for_add_sibling + (
+        'first-child', 'last-child', 'sorted-child')
+    _valid_pos_for_sorted_move = _valid_pos_for_sorted_add_sibling + (
+        'sorted-child',)
+
+    def _prepare_pos_var_for_move(self, pos):
+        return self._prepare_pos_var(
+            pos,
+            'move',
+            self._valid_pos_for_move,
+            self._valid_pos_for_sorted_move)
 
     def get_sorted_pos_queryset(self, siblings, newobj):
         """
         :returns: A queryset of the nodes that must be moved
         to the right. Called only for Node models with :attr:`node_order_by`
 
-        This function was taken from django-mptt (BSD licensed) by Jonathan
-        Buchanan:
-        http://code.google.com/p/django-mptt/source/browse/trunk/mptt/signals.py?spec=svn100&r=100#12
+        This function is based on _insertion_target_filters from django-mptt
+        (BSD licensed) by Jonathan Buchanan:
+        https://github.com/django-mptt/django-mptt/blob/0.3.0/mptt/signals.py
         """
 
         fields, filters = [], []
         for field in self.node_order_by:
             value = getattr(newobj, field)
-            filters.append(Q(*
-                [Q(**{f: v}) for f, v in fields] +
-                [Q(**{'%s__gt' % field: value})]))
+            filters.append(
+                Q(
+                    *[Q(**{f: v}) for f, v in fields] +
+                     [Q(**{'%s__gt' % field: value})]
+                )
+            )
             fields.append((field, value))
         return siblings.filter(reduce(operator.or_, filters))
 
     @classmethod
-    def get_annotated_list(cls, parent=None):
+    def get_annotated_list_qs(cls, qs):
+        """
+        Gets an annotated list from a queryset.
+        """
+        result, info = [], {}
+        start_depth, prev_depth = (None, None)
+        for node in qs:
+            depth = node.get_depth()
+            if start_depth is None:
+                start_depth = depth
+            open = (depth and (prev_depth is None or depth > prev_depth))
+            if prev_depth is not None and depth < prev_depth:
+                info['close'] = list(range(0, prev_depth - depth))
+            info = {'open': open, 'close': [], 'level': depth - start_depth}
+            result.append((node, info,))
+            prev_depth = depth
+        if start_depth and start_depth > 0:
+            info['close'] = list(range(0, prev_depth - start_depth + 1))
+        return result
+
+    @classmethod
+    def get_annotated_list(cls, parent=None, max_depth=None):
         """
         Gets an annotated list from a tree branch.
 
@@ -497,33 +597,18 @@ class Node(models.Model):
             The node whose descendants will be annotated. The node itself
             will be included in the list. If not given, the entire tree
             will be annotated.
+
+        :param max_depth:
+
+            Optionally limit to specified depth
         """
 
         result, info = [], {}
         start_depth, prev_depth = (None, None)
-
-        for node in cls.get_tree(parent):
-
-            depth = node.get_depth()
-
-            if start_depth is None:
-                start_depth = depth
-
-            open = (depth > prev_depth)
-
-            if depth < prev_depth:
-                info['close'] = range(0, prev_depth - depth)
-
-            info = {'open': open, 'close': [], 'level': depth - start_depth}
-
-            result.append((node, info,))
-
-            prev_depth = depth
-
-        if start_depth > 0:
-            info['close'] = range(0, prev_depth - start_depth + 1)
-
-        return result
+        qs = cls.get_tree(parent)
+        if max_depth:
+            qs = qs.filter(depth__lte=max_depth)
+        return cls.get_annotated_list_qs(qs)
 
     @classmethod
     def _get_serializable_model(cls):
@@ -534,36 +619,36 @@ class Node(models.Model):
 
         (this is a workaround for a bug in django)
         """
-        if DJANGO_VERSION >= (1, 1):
-            while cls._meta.proxy:
-                cls = cls._meta.proxy_for_model
-        return cls
+        current_class = cls
+        while current_class._meta.proxy:
+            current_class = current_class._meta.proxy_for_model
+        return current_class
 
     @classmethod
-    def get_database_engine(cls):
+    def _get_database_connection(cls, action):
+        return {
+            'read': connections[router.db_for_read(cls)],
+            'write': connections[router.db_for_write(cls)]
+        }[action]
+
+    @classmethod
+    def get_database_vendor(cls, action):
         """
-        Returns the supported database engine used by a treebeard model.
+        returns the supported database vendor used by a treebeard model when
+        performing read (select) or write (update, insert, delete) operations.
 
-        This will return the default database engine depending on the version
-        of Django. If you use something different, like a non-default database
-        in Django 1.2+, you need to override this method and return the correct
-        engine.
+        :param action:
 
-        :returns: postgresql, postgresql_psycopg2, mysql or sqlite3
+            `read` or `write`
+
+        :returns: postgresql, mysql or sqlite
         """
-        engine = None
-        if DJANGO_VERSION >= (1, 2):
-            try:
-                engine = settings.DATABASES['default']['ENGINE']
-            except AttributeError, KeyError:
-                engine = None
-            # the old style settings still work in Django 1.2+ if there is no
-            # DATABASES setting
-        if engine is None:
-            engine = settings.DATABASE_ENGINE
-        return engine.split('.')[-1]
+        return cls._get_database_connection(action).vendor
 
+    @classmethod
+    def _get_database_cursor(cls, action):
+        return cls._get_database_connection(action).cursor()
 
     class Meta:
-        "Abstract model."
+        """Abstract model."""
         abstract = True
